// -*- C++ -*-
// $RCSfile: cdebug.C,v $
// $Revision: 1.18 $
// $Author: langer $
// $Date: 2006-10-06 20:41:35 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

// Routines used to catch C++ signals and dump the current Python
// stack before bailing out.

#include <Python.h>
#include <string>
#include <signal.h>
#include <pthread.h>
#include <unistd.h>
#include <iostream>
#include "cdebug.h"

#include "common/threadstate.h"
#include "common/ooferror.h"

static PyObject *python_dumper;

static void handler(int sig) {
  std::string msg;
  if(sig == SIGBUS)
    msg = "Bus Error";
  else if(sig == SIGSEGV)
    msg = "Segmentation Fault";
  else if(sig == SIGINT)
    msg = "Interrupted";
  
  // No need to put the code between acquirePyLock() and
  // releasePyLock() in a try/catch block, because we're just going to
  // abort anyway.
  acquirePyLock();
  PyObject *args = Py_BuildValue("(s)", msg.c_str());
  PyObject *result = PyEval_CallObject(python_dumper, args);
  Py_DECREF(args);
  Py_DECREF(result);
  releasePyLock();
  std::cerr << "cdebug.C: " << msg << ": aborting" << std::endl;
  abort();
}

void initDebug(PyObject *pydump) {
  python_dumper = pydump;
  Py_XINCREF(python_dumper);
}

// installSignals and restoreSignals are called by the 'except'
// directive in typemaps.swg.

typedef void (*sighandler)(int);

// Signal handlers called due to internal segfaults (i.e. actual
// pointer-dereference problems, as opposed to "kill -11" from the
// command line) appear to be run on the segfaulting thread, resulting
// in the stack dump being correct.

// Experiments with test programs imply that successive "signal" calls
// override the main handler.  Since we actually only have one
// handler, and we want it to be in place whenever any thread is in
// C++ code, the right thing to do is to count up the handlers in a
// thread-safe way, and not restore until you're all the way out.

sighandler bushandler;
sighandler inthandler;
sighandler segvhandler;

static pthread_mutex_t signal_handler_count_lock=PTHREAD_MUTEX_INITIALIZER;
static int signal_handler_count=0;

void installSignals_() {
  pthread_mutex_lock(&signal_handler_count_lock);
  if (signal_handler_count==0) {
    bushandler = signal(SIGBUS, handler);
    inthandler = signal(SIGINT, handler);
    segvhandler = signal(SIGSEGV, handler);
  }
  signal_handler_count++;
  pthread_mutex_unlock(&signal_handler_count_lock);
}

void restoreSignals_() {
  pthread_mutex_lock(&signal_handler_count_lock);
  signal_handler_count--;
  if (signal_handler_count==0) {
    signal(SIGBUS, bushandler);
    signal(SIGINT, inthandler);
    signal(SIGSEGV, segvhandler);
  }
  pthread_mutex_unlock(&signal_handler_count_lock);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// The following functions are for testing the error handling
// facilities.  

void segfault(int delay) {
  // raise(SIGSEGV);
  sleep(delay);
  void (*f)() = 0;
  f();
}

void throwException() {
  throw ErrProgrammingError("Somebody made a mistake!", __FILE__, __LINE__);
}

// A C++ function that can be called from Python, and which calls a
// Python "function" that raises a Python exception.

void throwPythonException() {
  acquirePyLock();
  try {
    PyObject *result = PyObject_GetAttrString(Py_None, "say what?");
    if(!result)
      pythonErrorRelay();
    Py_XDECREF(result);
  }
  catch(...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}

// A C++ function that can be called from Python, and which calls a
// Python function that throws a C++ exception.

void throwPythonCException() {
  acquirePyLock();
  try {
    PyObject *module = PyImport_ImportModule("oof2.SWIG.common.ooferror");
    PyObject *func = PyObject_GetAttrString(module, "pytesterror");
    Py_XDECREF(module);
    PyObject *args = Py_BuildValue("()");
    PyObject *result = PyEval_CallObject(func, args);
    Py_XDECREF(args);
    if(!result)
      pythonErrorRelay();
    Py_XDECREF(result);
  }
  catch (...) {
    releasePyLock();
    throw;
  }
  releasePyLock();
}
