// -*- C++ -*-
// $RCSfile: lock.C,v $
// $Revision: 1.17 $
// $Author: reida $
// $Date: 2006-08-01 21:02:09 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */


#include <oofconfig.h>
#include <pthread.h>
#include <iostream>
#include <errno.h>
#include "lock.h"
#include "ooferror.h"
#include "threadstate.h"

static bool enabled = true;

void disable_all() {
  enabled = false;
}

void enable_all() {
  enabled = true;
}

Lock::Lock() {
  if(enabled) 
    pthread_mutex_init(&lock, NULL);
}

Lock::~Lock() {
  if(enabled)
    pthread_mutex_destroy(&lock);
}


void Lock::acquire() {
  if(enabled) {
    if (findThreadNumber()==0) {
      std::cerr << "Lock::acquire for " << this << " called on main thread." << std::endl;
      throw ErrProgrammingError("Lock error.", __FILE__, __LINE__);
    }
    pthread_mutex_lock(&lock);
  }
}

void SLock::acquire() {
  if(enabled) {
    pthread_mutex_lock(&lock);
  }
}

void Lock::release() {
  if(enabled)
    pthread_mutex_unlock(&lock);
}



RWLock::RWLock()
  : r(0), w(0), p(0)
{
  if(enabled) {
    pthread_mutex_init(&local_lock, NULL);
    pthread_cond_init(&rw_zero, NULL);
  }
}

RWLock::~RWLock() {
  if(enabled) {
    pthread_mutex_destroy(&local_lock);
    pthread_cond_destroy(&rw_zero);
  }
}


// Before you can write, you have to wait until there are no 
// readers or writers.  
void RWLock::write_acquire() {
  if(enabled) {
    if (findThreadNumber()==0) {
      std::cerr << "RWLock::write_acquire called on main thread." << std::endl;
      throw ErrProgrammingError("RWLock write error.", __FILE__, __LINE__);
    }
    pthread_mutex_lock(&local_lock);
    while (r>0 || w>0 || p>0) {
      pthread_cond_wait(&rw_zero, &local_lock);
    }
    w=1;
    pthread_mutex_unlock(&local_lock);
  }
}


void RWLock::write_release() {
  if(enabled) {
    pthread_mutex_lock(&local_lock);
    w = 0;
    p = 0;
    pthread_cond_broadcast(&rw_zero);  // r==0 at this point.
    pthread_mutex_unlock(&local_lock);
  }
}


// You cannot read if anybody is writing.  Other readers are OK.
void RWLock::read_acquire() {
  if(enabled) {
    if (findThreadNumber()==0) {
      std::cerr << "RWLock::read_acquire called on main thread." << std::endl;
      throw ErrProgrammingError("RWLock read error.", __FILE__, __LINE__);
    }
    pthread_mutex_lock(&local_lock);
    while (w>0) {
      pthread_cond_wait(&rw_zero, &local_lock);
    }
    r++;
    pthread_mutex_unlock(&local_lock);
  }
}

void RWLock::read_release() {
  if(enabled) {
    pthread_mutex_lock(&local_lock);
    r--;
    if (r==0)  // w==0 is already true here.
      pthread_cond_broadcast(&rw_zero);
    pthread_mutex_unlock(&local_lock);
  }
}

void RWLock::write_pause() {
  if(enabled) {
    bool ok = true;
    pthread_mutex_lock(&local_lock);
    if(w == 0) ok = false;
    p = 1;
    w = 0;
    pthread_cond_broadcast(&rw_zero);
    pthread_mutex_unlock(&local_lock);
    if(!ok)
      throw ErrProgrammingError("Called RWLock::write_pause() when not writing",
				__FILE__, __LINE__);
  }
}

void RWLock::write_resume() {
  if(enabled) {
    pthread_mutex_lock(&local_lock);
    while (r>0) {
      //std::cerr << "***** Waiting in RWLock::write_resume" << std::endl;
      pthread_cond_wait(&rw_zero, &local_lock);
    }
    p = 0;
    w = 1;
    pthread_mutex_unlock(&local_lock);
  }
}




// SRWLock functions -- same as RWLock except for error-checking.


// Before you can write, you have to wait until there are no 
// readers or writers.  
void SRWLock::write_acquire() {
  if(enabled) {
    pthread_mutex_lock(&local_lock);
    while (r>0 || w>0 || p>0) {
      pthread_cond_wait(&rw_zero, &local_lock);
    }
    w=1;
    pthread_mutex_unlock(&local_lock);
  }
}


// You cannot read if anybody is writing.  Other readers are OK.
void SRWLock::read_acquire() {
  if(enabled) {
    pthread_mutex_lock(&local_lock);
    while (w>0) {
      pthread_cond_wait(&rw_zero, &local_lock);
    }
    r++;
    pthread_mutex_unlock(&local_lock);
  }
}


KeyHolder::KeyHolder(Lock &some_lock) : lock(&some_lock) {
  if(enabled)
    lock->acquire();
}


KeyHolder::~KeyHolder()
{
  if(enabled)
    lock->release();
}
