# -*- python -*-
# $RCSfile: skeletoninfo.py,v $
# $Revision: 1.32 $
# $Author: rlua $
# $Date: 2007-03-15 21:48:50 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import switchboard
from oof2.SWIG.common import timestamp
from oof2.common import debug
from oof2.common import primitives
from oof2.common import ringbuffer
from oof2.common import toolbox
from oof2.common.IO import oofmenu
from oof2.common.IO import parameter
from oof2.common.IO import reporter
from oof2.common.IO import xmlmenudump
from oof2.engine import deputy
from oof2.engine import skeletonelement
from oof2.engine import skeletonnode
from oof2.engine import skeletonsegment

from oof2.common import parallel_enable
if parallel_enable.enabled():
    from oof2.SWIG.common import mpitools
    from oof2.engine.IO import skeletonIPC

SkeletonElement = skeletonelement.SkeletonElement
SkeletonSegment = skeletonsegment.SkeletonSegment
SkeletonNode = skeletonnode.SkeletonNode

class SkeletonQueryContainer:
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context
        self.skeleton = context.getObject()
        self.object = None
        self.position = None
        self.targetname = None

    def set(self, object=None, targetname=None, position=None):
        self.object = object
        self.targetname = targetname
        self.position = position
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp

    def clone(self):
        krusty = SkeletonQueryContainer(self.context)
        krusty.skeleton = self.skeleton
        krusty.object = self.object
        krusty.targetname = self.targetname
        krusty.position = self.position
        krusty.timestamp.increment()
        return krusty

    def clearable(self):
        return not not self.object

class SkeletonPeekContainer(SkeletonQueryContainer):
    def __init__(self, context):
        self.timestamp = timestamp.TimeStamp()
        self.context = context
        self.skeleton = context.getObject()
        self.objects = {"Element":None, "Segment":None, "Node":None}

    def assignObject(self, object, objtype):
        self.objects[objtype] = object
        self.timestamp.increment()

    def removeObject(self, objtype):
        self.objects[objtype] = None
        self.timestamp.increment()

    def reset(self):
        self.objects = {"Element":None, "Segment":None, "Node":None}
        self.timestamp.increment()

    def getTimeStamp(self):
        return self.timestamp

##################################################################
            
class SkeletonInfoToolbox(toolbox.Toolbox):
    def __init__(self, gfxwindow):
        toolbox.Toolbox.__init__(self, 'Skeleton_Info', gfxwindow)
        self.whoset = ('Skeleton',)
        self.querier = None
        self.peeker = None        
        self.records = ringbuffer.RingBuffer(49)

        self.sbcallbacks = [
            # Looks for a skeleton on the gfx window.
            switchboard.requestCallback((self.gfxwindow(), "layers changed"),
                                        self.newLayers),
            # Looks for a skeleton modification.
            switchboard.requestCallback(('who changed', 'Skeleton'),
                                        self.skelChanged)
            ]

    def close(self):
        map(switchboard.removeCallback, self.sbcallbacks)

    def getSkeletonContext(self):
        return self.gfxwindow().topwho(*self.whoset)

    def makeMenu(self, menu):
        self.menu = menu
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryElement',
            callback=self.queryElem,
            params=[primitives.PointParameter('position', tip='Target point.')],
            help="Query the element closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/query_skel_elem.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'QuerySegment',
            callback=self.querySgmt,
            params=[primitives.PointParameter('position', tip='Target point.')],
            help="Query the segment closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/query_skel_sgmt.xml')
            ))
        menu.addItem(oofmenu.OOFMenuItem(
            'QueryNode',
            callback=self.queryNode,
            params=[primitives.PointParameter('position', tip='Target point.')],
            help="Query the node closest to the given point.",
            discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/menu/query_skel_node.xml')
            ))

    def activate(self):
        # Called when this toolbox has been brought to the front.
        skelcontext = self.getSkeletonContext()
        if skelcontext is not None:
            if not self.querier:
                self.querier = SkeletonQueryContainer(skelcontext)
            if not self.peeker:
                self.peeker = SkeletonPeekContainer(skelcontext)
        else:
            self.querier = None
            self.peeker = None

    def newLayers(self):
        skelcontext = self.getSkeletonContext()
        if skelcontext is not None:
            if self.querier and \
               skelcontext is not self.querier.context:
                self.resetRecords()
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)
                self.redraw()
            else:
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)
        switchboard.notify( (self, "new layers") )

    def skelChanged(self, context):
        skelcontext = self.getSkeletonContext()
        if not skelcontext:
            return  # No skeleton on the gfx window.
        if context is skelcontext and \
           isinstance(context.getObject(), deputy.DeputySkeleton):
            if self.querier:
                if context.getObject().sheriffSkeleton() is \
                   self.querier.skeleton:
                    # It should be node-only modification.
                    # Thus, it is not necessary to dump all the records but
                    # needs to reflect new node positions, if applicable.
                    self.querysignal()
                else:
                    self.resetRecords()
                    self.querier = SkeletonQueryContainer(skelcontext)
                    self.peeker = SkeletonPeekContainer(skelcontext)
                    self.querysignal()                    
            else:
                self.querier = SkeletonQueryContainer(skelcontext)
                self.peeker = SkeletonPeekContainer(skelcontext)
        else:
            # The skeleton has been modified.
            # Thus, the query records should be reset, since "prev",
            # "next" would be meaningless after the modifications such as
            # "Refine" or "Rationalize".
            self.resetRecords()
            self.querier = SkeletonQueryContainer(skelcontext)
            self.peeker = SkeletonPeekContainer(skelcontext)
            self.querysignal()
        
    def resetRecords(self):
        self.records.clear()

    def queryElem(self, menuitem, position):
        context = self.getSkeletonContext()
        if not context:
            return
        skeleton = context.getObject()

        if parallel_enable.enabled():
            skeletonIPC.smenu.Skel_Info_Query(targetname="Element",
                                              position=position,skeleton=context.path())
            if mpitools.Rank()>0:
                return

        elem = skeleton.enclosingElement(position)
        if not elem:
            reporter.report("Try to click ON an Element, dude.")
        else:
            self.finishQuery(elem, "Element", position)

    def querySgmt(self, menuitem, position):
        context = self.getSkeletonContext()
        if not context:
            return
        skeleton = context.getObject()

        if parallel_enable.enabled():
            skeletonIPC.smenu.Skel_Info_Query(targetname="Segment",
                                              position=position,skeleton=context.path())
            if mpitools.Rank()>0:
                return

        sgmt = skeleton.nearestSgmt(position)
        if not sgmt:
            reporter.report("Avoid clicking OUTSIDE of a skeleton.")
        else:
            self.finishQuery(sgmt, "Segment", position)

    def queryNode(self, menuitem, position):            
        context = self.getSkeletonContext()
        if not context:
            return
        skeleton = context.getObject()

        if parallel_enable.enabled():
            skeletonIPC.smenu.Skel_Info_Query(targetname="Node",
                                              position=position,skeleton=context.path())
            if mpitools.Rank()>0:
                return

        node = skeleton.nearestNode(position)
        if not node:
            reporter.report("Avoid clicking OUTSIDE of a skeleton.")
        else:
            self.finishQuery(node, "Node", position)

    def finishQuery(self, object, targetname, position):
        self.querier.set(object=object, targetname=targetname,
                         position=position)
        self.records.push(self.querier.clone())
        self.peeker.reset()
        self.querysignal()
        
    def querysignal(self):
        switchboard.notify((self.gfxwindow(),"query skeleton"))
        self.redraw()

    def clearQuerier(self):
        if self.querier:
            self.querier.set()
            self.peeker.reset()
            self.redraw()

    def redraw(self):
        switchboard.notify("redraw")
        
    def prev_able(self):
        return not self.records.atBottom()

    def next_able(self):
        return not self.records.atTop()

    tip="Get information about Skeleton components."
    discussion="""<para>
    Get information about &skel; components, based on mouse input.
    </para>"""
    
toolbox.registerToolboxClass(SkeletonInfoToolbox, ordering=2.0)
