// -*- C++ -*-
// $RCSfile: piezoelectricity.C,v $
// $Revision: 1.40 $
// $Author: langer $
// $Date: 2007-05-10 21:22:19 $


/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>


#include "common/coord.h"
#include "engine/IO/propertyoutput.h"
#include "engine/celectricfield.h"
#include "engine/cstrain.h"
#include "engine/csubproblem.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/equation.h"
#include "engine/flux.h"
#include "engine/material.h"
#include "engine/matvec.h"
#include "engine/property/elasticity/cijkl.h"
#include "engine/property/elasticity/elasticity.h"
#include "engine/property/orientation/orientation.h"
#include "engine/property/piezoelectricity/piezoelectricity.h"
#include "engine/rank3tensor.h"

PiezoElectricity::PiezoElectricity(PyObject *reg, 
				   const std::string &nm)
  : Property(nm,reg)
{
  
  stress_flux=dynamic_cast<SymmetricTensorFlux*>(Flux::getFlux("Stress"));
  displacement = dynamic_cast<TwoVectorField*>(Field::getField("Displacement"));
  voltage=dynamic_cast<ScalarField*>(Field::getField("Voltage"));
  total_polarization = dynamic_cast<VectorFlux*>(Flux::getFlux("Total_Polarization"));
}


void PiezoElectricity::precompute(FEMesh*) {

}

int PiezoElectricity::integration_order(const CSubProblem *subproblem, 
					const Element *el) const {
  if(displacement->in_plane(subproblem))
    return el->dshapefun_degree();
  return el->shapefun_degree();
}


void PiezoElectricity::fluxmatrix(const FEMesh *mesh,
				  const Element *element,
				  const ElementFuncNodeIterator &nu, 
				  const Flux *flux, FluxData *fluxdata,
				  const MasterPosition &pos) const
{
  const Cijkl modulus = elasticity->cijkl(mesh, element, pos);
  const Rank3Tensor dijkLocal = dijk(mesh, element, pos);
  const Rank3Tensor eijk = modulus*dijkLocal;
  //shape functions and their derivatives
  double sf = nu.shapefunction(pos);
  double dsf0 = nu.dshapefunction(0, pos);
  double dsf1 = nu.dshapefunction(1, pos);
  

  //contributions to the stress flux
  if(*flux == *stress_flux) {
    SymTensorIterator kl;	// reuse inner loop iterator
    for(SymTensorIterator ij; !ij.end(); ++ij) {
      double &mtx_el = fluxdata->matrix_element(ij, voltage, nu);
      // Strain tensor components.
      for( ; !kl.end(); ++kl) {
	double ca = modulus(ij, kl)*(dijkLocal(0,kl)*dsf0 + 
				     dijkLocal(1,kl)*dsf1);
	if(!voltage->in_plane(mesh)) {
	  ca += modulus(ij, kl)*dijkLocal(2,kl)*sf;
	}
	if(kl.diagonal()) {
	  mtx_el += ca;
	}
	else {
	  mtx_el += 2.0*ca;
	}
      }
      kl.reset();
    }
  }
  
  //contributions to total_polarization flux
    
  // Loop over flux components.  Loop over all components, even if
  // the flux is in-plane, because the out-of-plane components of
  // the flux matrix are used to construct the constraint equation.

  if(*flux == *total_polarization)
    for(VectorFieldIterator i; !i.end(); ++i) {
      // in-plane displacement gradient contributions
      int ii = i.integer();
      fluxdata->matrix_element(i, displacement, nu) +=
	eijk(ii, 0, 0) * dsf0 + 
	eijk(ii, 1, 0) * dsf1 + 
	eijk(ii, 0, 1) * dsf0 + 
	eijk(ii, 1, 1) * dsf1 +
	eijk(ii, 0, 2) * dsf0 + 
	eijk(ii, 1, 2) * dsf1;
      // out-of-plane strain contribution
      if(!displacement->in_plane(mesh)) {
	if (ii == 2) {
	  fluxdata->matrix_element(i, displacement->out_of_plane(), nu) += 
	    eijk(2, 2, 0) * sf +
	    eijk(2, 2, 1) * sf +
	    eijk(2, 2, 2) * sf;
	}
      }
    }   // Loop over all tensor components.
    
  
  // TODO: Should check for, and complain about, an unexpected flux
  // here, if it comes to that.
  
}



void PiezoElectricity::output(const FEMesh *mesh,
			      const Element *element,
			      const PropertyOutput *output,
			      const MasterPosition &pos,
			      OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Strain") {
    std::string stype = output->getRegisteredParamName("type");
    SymmMatrix3 *sdata = dynamic_cast<SymmMatrix3*>(data);
    VECTOR_D E_field(3,0);
    findElectricField(mesh, element, pos, E_field);
    if(stype == "Piezoelectric")
      *sdata += dijk(mesh, element, pos)*E_field;
    else if(stype == "Elastic")
      *sdata -= dijk(mesh, element, pos)*E_field;
  }

  if(outputname == "Energy") {
    std::string etype = output->getEnumParam("etype");
    //Energies
    if(etype == "Total" || etype == "Elastic" || etype == "Electric") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      SymmMatrix piezoelectricStrain(3);
      const Cijkl modulus = elasticity->cijkl(mesh, element, pos); //returns lab reference system stiffness
      
      VECTOR_D E_field(3,0.0);
      findElectricField(mesh, element, pos, E_field);
      piezoelectricStrain = dijk(mesh, element, pos)*E_field;
      SymmMatrix electricstress(modulus*piezoelectricStrain);
      SymmMatrix strain(3);
      findGeometricStrain(mesh, element, pos, strain);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += electricstress(i,i)*strain(i,i);
	int j = (i+1)%3;
	e += 2*electricstress(i,j)*strain(i,j);
      }
      if(etype == "Total")
	*edata += -e;      
      else if (etype == "Elastic" || etype == "Electric")
	*edata += -0.5*e; 
    } //different types of energies end here
    
  }//energy output ends here
}

IsotropicPiezoElectricity::IsotropicPiezoElectricity(PyObject *registry,
						     const std::string &name, 
						     double d)
  : PiezoElectricity(registry, name),
    _dijkValue(d)
{

}


void IsotropicPiezoElectricity::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  
}

void IsotropicPiezoElectricity::precompute(FEMesh *mesh) {
  PiezoElectricity::precompute(mesh);
  _dijkLab(0,0,0) = _dijkLab(1,1,1)
    = _dijkLab(2,2,2) = _dijkValue;
}

AnisotropicPiezoElectricity::AnisotropicPiezoElectricity(PyObject *registry, 
							 const std::string &nm, 
							 Rank3Tensor *dijkTensor)
  : PiezoElectricity(registry, nm),
    _dijkValue(*dijkTensor)
{

}
void AnisotropicPiezoElectricity::cross_reference(Material *mat) {
  // find out which property is the elasticity
  elasticity = dynamic_cast<Elasticity*>(mat->fetchProperty("Elasticity"));
  orientation = dynamic_cast<OrientationPropBase*>
    (mat->fetchProperty("Orientation"));
}

void AnisotropicPiezoElectricity::precompute(FEMesh*) {
  if(orientation->constant_in_space())
    _dijkLab = _dijkValue.transform(orientation->orientation());
}

const Rank3Tensor
AnisotropicPiezoElectricity::dijk(const FEMesh *mesh, const Element *el,
				  const MasterPosition &pos) const
{
  if(orientation->constant_in_space())
    return _dijkLab;
  return _dijkValue.transform(orientation->orientation(mesh, el, pos));
}
